#include "AStarPlanner.h"

using namespace std;

AStarPlanner::AStarPlanner(VirtualRobot::RobotPtr robot, VirtualRobot::SceneObjectSetPtr obstacles, float cellSize)
 : Planner2D(robot, obstacles), cellSize(cellSize)
{
}

void AStarPlanner::createUniformGrid()
{
    //+1 for explicit rounding up
    size_t cols = (sceneBoundsMax.x() - sceneBoundsMin.x()) / cellSize + 1;
    size_t rows = (sceneBoundsMax.y() - sceneBoundsMin.y()) / cellSize + 1;

    // build objects in grid
    for (size_t r = 0; r < rows; r++) {
        grid.push_back(std::vector<NodePtr>(cols));
        for (size_t c = 0; c < cols; c++) {
            Eigen::Vector2f pos;
            pos.x() = sceneBoundsMin.x() + c * cellSize + cellSize / 2;
            pos.y() = sceneBoundsMin.y() + r * cellSize + cellSize / 2;
            grid[r][c] = NodePtr(new Node(pos));
        }
    }

    //init valid successors
    for (size_t r = 0; r < rows; r++) {
        for (size_t c = 0; c < cols; c++) {
            NodePtr candidate = grid[r][c];
            if (!fulfillsConstraints(candidate)) continue;
            //add the valid node as successor to all its neighbors
            for (int nR = -1; nR <= 1; nR++) {
                for (int nC = -1; nC <= 1; nC++) {
                    const int neighborIndexX = c + nC;
                    const int neighborIndexY = r + nR;
                    if (neighborIndexX < 0 || neighborIndexY < 0 || (nR == 0 && nC == 0)) continue;

                    grid[neighborIndexY][neighborIndexX]->successors.push_back(candidate);
                }
            }
        }
    }

}

bool AStarPlanner::fulfillsConstraints(NodePtr n)
{
    if (!obstacles) return true;
    robotCollisionModel->setGlobalPose(positionToMatrix4f(n->position));

    // check for collisions
    if (!obstacles) return true;
    robotCollisionModel->setGlobalPose(positionToMatrix4f(n->position));
    Eigen::Vector3f P1, P2;
    int id1, id2;
    for (size_t i = 0; i < obstacles->getSize(); i++) {
        float dist = VirtualRobot::CollisionChecker::getGlobalCollisionChecker()->calculateDistance(robotCollisionModel, obstacles->getSceneObject(i)->getCollisionModel(), P1, P2, &id1, &id2);
        if (dist <= cellSize / 2) return false;
    }
    return n->position.x() >= sceneBoundsMin.x() && n->position.x() <= sceneBoundsMax.x() &&
           n->position.y() >= sceneBoundsMin.y() && n->position.y() <= sceneBoundsMax.y();
}

NodePtr AStarPlanner::closestNode(Eigen::Vector2f v)
{
    size_t r = (v.y() - cellSize / 2 - sceneBoundsMin.y()) / cellSize;
    size_t c = (v.x() - cellSize / 2 - sceneBoundsMin.x()) / cellSize;
    return grid[r][c];
}

float AStarPlanner::heuristic(NodePtr n1, NodePtr n2)
{
    return (n1->position - n2->position).norm();
}

std::vector<Eigen::Vector2f> AStarPlanner::plan(Eigen::Vector2f start, Eigen::Vector2f goal)
{
    grid.clear();
    std::vector<Eigen::Vector2f> result;
    if (!robot || !robot->hasRobotNode(robotColModelName) || !robot->getRobotNode(robotColModelName)->getCollisionModel())
    {
        cout << "No collision model with name " << robotColModelName << " found..." << endl;
        return result;
    }
    robotCollisionModel = robot->getRobotNode(robotColModelName)->getCollisionModel()->clone();
    createUniformGrid();
    NodePtr v_start = closestNode(start);
    NodePtr v_goal = closestNode(goal);

    // Open set O: nodes to visit
    std::set<NodePtr> openSet;
    // Closed set C: nodes which have been expanded/visited
    std::set<NodePtr> closedSet;
    // Accumulated costs g(v)
    std::map<NodePtr, float> g;
    // Heuristic h(v)
    std::map<NodePtr, float> h;
    // Expected cost f(v) = g(v) + h(v)

    ///////////////////////////////////////////////
    // INSERT CODE HERE
    ///////////////////////////////////////////////

    //found solution, now retrieve path from goal to start
    if (v_goal)
        result = v_goal->traversePredecessors();
    //since the graph was traversed from goal to start, we have to reverse the order
    std::reverse(result.begin(), result.end());

    return result;
}
